SEXP
	Matrix_DimNamesSym,
	Matrix_DimSym,
	Matrix_LSym,
	Matrix_QSym,
	Matrix_RSym,
	Matrix_TSym,
	Matrix_USym,
	Matrix_VSym,
	Matrix_betaSym,
	Matrix_diagSym,
	Matrix_factorsSym,
	Matrix_iSym,
	Matrix_jSym,
	Matrix_lengthSym,
	Matrix_marginSym,
	Matrix_pSym,
	Matrix_permSym,
	Matrix_qSym,
	Matrix_sdSym,
	Matrix_uploSym,
	Matrix_xSym;
