#ifndef MATRIX_VERSION_H
#define MATRIX_VERSION_H

/* (version)_{10} = (major minor patch)_{256} */
#define MATRIX_PACKAGE_VERSION 67334
#define MATRIX_PACKAGE_MAJOR 1
#define MATRIX_PACKAGE_MINOR 7
#define MATRIX_PACKAGE_PATCH 6

#define MATRIX_ABI_VERSION 2

/* (version)_{10} = (major minor patch)_{256} */
#define MATRIX_SUITESPARSE_VERSION 460288
#define MATRIX_SUITESPARSE_MAJOR 7
#define MATRIX_SUITESPARSE_MINOR 6
#define MATRIX_SUITESPARSE_PATCH 0

#endif /* MATRIX_VERSION_H */
