#ifndef R_MATRIX_MATRIX_H
#define R_MATRIX_MATRIX_H

#include "version.h"
#include "cholmod.h"

#ifndef R_MATRIX_NO_CHOLMOD_UTILS
# include "cholmod-utils.h"
#endif

#ifndef R_MATRIX_NO_ALLOCA
# include "alloca.h"
#endif

#ifndef R_MATRIX_NO_REMAP
# include "remap.h"
#endif

#endif /* R_MATRIX_MATRIX_H */
